
// Standardne pre MS win aplikacie
#include <windows.h>

// GLEW kniznica. pre spravu novsich rozsireni
#include "glew.h"
// GL Utility Toolkit (GLUT), v ramci toho sa includuje aj openGL hlavickovy subor
#include "glut.h"

#include <stdio.h>
#include <math.h>

#include "shaders.h"
#include "textures.h"
#include "geometry.h"
#include "fonts.h"

// Globalne premenne
GLboolean g_isAnim = true;
GLboolean g_isMultisample = true;
GLboolean g_isTexture = false;
float sun_rotation = 0.0f;

// Globalne premenne pre prezeranie sceny
GLdouble vzd, old_vzd, fi, xi, old_fi, old_xi;
GLint left_mouse_down_x, left_mouse_down_y;
GLint right_mouse_down_y;
GLboolean right_down = false, left_down = false;

GLuint g_window_height, g_window_width;

// id textur
GLuint tex_sun, tex_earth, tex_lava, tex_phobos, tex_skybox, tex_point_sprite;

// id shaderov
GLuint shaders_skybox, shaders_perpixel_phong, shaders_envmap;

// id naposledny kliknuteho objektu
GLuint g_id_selected_object = -1;

// casticovy system
#define NUM_PARTICLES 1000
struct Particle
{
    float position[3];
    float velocity[3];
    float color[4];
    float life;
};
Particle g_particle_system[NUM_PARTICLES];


float getRandomMinMax(float fMin, float fMax)
{
    float fRandNum = (float)rand () / RAND_MAX;
    return fMin + (fMax - fMin) * fRandNum;
}

void getRandomVector(float* vector)
{
    vector[2] = getRandomMinMax( 0.4f, 1.0f );
    float radius = (float)sqrt(1 - vector[2] * vector[2]);
    float t = getRandomMinMax( -PI, PI );
    vector[0] = (float)cosf(t) * radius;
    vector[1] = (float)sinf(t) * radius;
}

void initParticles()
{
    for( int i = 0; i < NUM_PARTICLES; i++ )
    {
        // emitter of all particles is point [0,0,0]
        g_particle_system[i].position[0] = 0;
        g_particle_system[i].position[1] = 0;
        g_particle_system[i].position[2] = 0;
        getRandomVector(g_particle_system[i].velocity);                 
        g_particle_system[i].color[0] = getRandomMinMax( 0.0f, 1.0f );
        g_particle_system[i].color[1] = getRandomMinMax( 0.0f, 1.0f );
        g_particle_system[i].color[2] = getRandomMinMax( 0.0f, 1.0f );
        g_particle_system[i].color[3] = getRandomMinMax( 0.0f, 0.4f );
        g_particle_system[i].life = getRandomMinMax( 1.0f, 3.0f );
    }
}

void updateParticles()
{
    static double dLastFrameTime = timeGetTime();
    double dCurrenFrameTime = timeGetTime();
    double dElpasedFrameTime = (float)((dCurrenFrameTime - dLastFrameTime) * 0.001);
    dLastFrameTime = dCurrenFrameTime;
    for( int i = 0; i < NUM_PARTICLES; i++ )
    {
        g_particle_system[i].position[0] += (float)dElpasedFrameTime * g_particle_system[i].velocity[0];
        g_particle_system[i].position[1] += (float)dElpasedFrameTime * g_particle_system[i].velocity[1];
        g_particle_system[i].position[2] += (float)dElpasedFrameTime * g_particle_system[i].velocity[2];
        g_particle_system[i].life -= dElpasedFrameTime;
        if (g_particle_system[i].life < 0 || g_particle_system[i].position[2] < 0)
        {
            g_particle_system[i].position[0] = 0;
            g_particle_system[i].position[1] = 0;
            g_particle_system[i].position[2] = 0;
            getRandomVector(g_particle_system[i].velocity);                 
            g_particle_system[i].color[0] = getRandomMinMax( 0.0f, 1.0f );
            g_particle_system[i].color[1] = getRandomMinMax( 0.0f, 1.0f );
            g_particle_system[i].color[2] = getRandomMinMax( 0.0f, 1.0f );
            g_particle_system[i].color[3] = getRandomMinMax( 0.0f, 0.4f );
            g_particle_system[i].life = getRandomMinMax( 1.0f, 3.0f );
        }
    }
}

void renderParticles()
{
    glEnable( GL_BLEND );
    glBlendFunc( GL_SRC_ALPHA, GL_ONE );
    glDepthMask(GL_FALSE);

    float quadratic[] =  { 1.0f, 0.0f, 0.01f };
    glPointParameterfvARB( GL_POINT_DISTANCE_ATTENUATION_ARB, quadratic );
    glPointSize(50);
    glPointParameterfARB( GL_POINT_FADE_THRESHOLD_SIZE_ARB, 60.0f );
    glPointParameterfARB( GL_POINT_SIZE_MIN_ARB, 1.0f );
    glPointParameterfARB( GL_POINT_SIZE_MAX_ARB, 50.0f );

    glBindTexture(GL_TEXTURE_2D, tex_point_sprite);
    glTexEnvf( GL_POINT_SPRITE_ARB, GL_COORD_REPLACE_ARB, GL_TRUE );
    glEnable( GL_POINT_SPRITE_ARB );
    if (g_isTexture)
        glEnable (GL_TEXTURE_2D);
    else
        glDisable (GL_TEXTURE_2D);
    glBegin( GL_POINTS );
    for( int i = 0; i < NUM_PARTICLES; i++ )
    {
        glColor4fv(g_particle_system[i].color);
        glVertex3fv(g_particle_system[i].position);
    }
    glEnd();
    glDisable( GL_POINT_SPRITE_ARB );

    glDepthMask(GL_TRUE);
    glDisable( GL_BLEND );
}


void initShaders()
{
    // nacitanie, kompilacia a zlinkovanie shaderov
    shaders_skybox = loadProgram("shaders/skybox.vert", "shaders/skybox.frag");
    SetShaderUniform1i(shaders_skybox, "cube_map", 0);

    shaders_perpixel_phong = loadProgram("shaders/perpixel_phong.vert", "shaders/perpixel_phong.frag");
    SetShaderUniform1i(shaders_perpixel_phong, "color_texture", 0);

    shaders_envmap = loadProgram("shaders/perpixel_envmap.vert", "shaders/perpixel_envmap.frag");
    SetShaderUniform1i(shaders_envmap, "color_texture", 0);
    SetShaderUniform1i(shaders_envmap, "cubemap_texture", 1);

    glActiveTexture(GL_TEXTURE0);
}

// inicializacia OpenGL textur
void initTextures()
{
    // ************* prva textura, zem *************
    tex_earth = LoadTexture("textures/earth.jpg");

    // ************* druha textura, slnko *************
    tex_sun = LoadTexture("textures/sun.jpg");

    // ************* tretia textura, jupiter *************
    tex_lava = LoadTexture("textures/lava.png");

    // ************* stvrta textura, phobos *************
    tex_phobos = LoadTexture("textures/phobos.jpg");

    // ************* piata textura, phobos *************
    tex_skybox = LoadCubeTexture("skybox/snow_positive_x.jpg", "skybox/snow_negative_x.jpg",
                                 "skybox/snow_positive_y.jpg", "skybox/snow_negative_y.jpg",
                                 "skybox/snow_positive_z.jpg", "skybox/snow_negative_z.jpg");

    // ************* textura pre casticovy system *************
    tex_point_sprite = LoadTexture("textures/particle.png");
}

// Open GL Inicializacia
bool init(void)
{
    // Cierne pozadie
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    // Nastav z-buffer testovanie
    glClearDepth(1.0f);
    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    // nastav stencil buffer
    glClearStencil(0);

    // nastav zmiesavaciu funkciu
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

    // nastav vhodne mapovanie textur
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

    // initializuj kameru
    fi = 0.7f; xi = 0.7f; vzd = 10.0f;

    // initializuje svetlo
    GLfloat light_ambient[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    GLfloat light_diffuse[] = { 0.9f, 0.9f, 0.9f, 1.0f };
    GLfloat light_specular[] = { 0.5f, 0.5f, 0.5f, 1.0f };
    glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular);
    glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 1.0);
    glEnable(GL_LIGHT0);

    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    GLint samples;
    glGetIntegerv(GL_SAMPLES, &samples);
    printf("Using context with %ix multisampling.\n", samples);
    if (g_isMultisample)
        glEnable(GL_MULTISAMPLE);

	return true;
}

void renderScene(bool selection)
{
    if (selection)
    {
        glInitNames();
        glPushName(-1);
    }

    // Nastav modelview maticu (transformaciu) na jednotkovu maticu
    glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();									

    // Nastav view matice pre pohlad na svetovy suradnicovy system
    GLfloat camera_pos[4];
    camera_pos[0] = vzd*cos(fi)*cos(xi);
	camera_pos[1] = vzd*sin(fi)*cos(xi);
	camera_pos[2] = vzd*sin(xi);
    camera_pos[3] = 1;
	gluLookAt(camera_pos[0], camera_pos[1], camera_pos[2], 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f );
    float view_matrix[16];
    glGetFloatv(GL_MODELVIEW_MATRIX, view_matrix);

    // vykresli suradnicove osi svetoveho systemu roznymi farbami
    if (!selection)
    {
        glLineWidth(1);
        glBegin( GL_LINES );
        glColor3f(1.0f, 0.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 100.0f, 0.0f, 0.0f );
        glColor3f(0.0f, 1.0f, 0.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 100.0f, 0.0f );
        glColor3f(0.0f, 0.0f, 1.0f );glVertex3f( 0.0f, 0.0f,0.0f );glVertex3f( 0.0f, 0.0f, 100.0f );
        glEnd();
    }

    if (!selection)
    {
        // vykresli skybox
        // skybox je dany vo svetovych suradniciach, akurat jeho stred je stale v bode, kdeje kamera
        glUseProgram(shaders_skybox);
        glActiveTexture(GL_TEXTURE0);
        glBindTexture(GL_TEXTURE_CUBE_MAP, tex_skybox);
        glPushMatrix();
        glScalef(500, 500, 500);
        renderCubeVBO();
        glPopMatrix();
        glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
        glUseProgram(0);
    }

    // using per-pixel phong
    glUseProgram(shaders_perpixel_phong);
    if (g_isTexture)
        SetShaderUniform1i(shaders_perpixel_phong, "texturing_enabled", 1);
    else
        SetShaderUniform1i(shaders_perpixel_phong, "texturing_enabled", 0);

    // vykresli slnko
    SetShaderUniform1i(shaders_perpixel_phong, "lighting_enabled", 0);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex_sun);
    glPushMatrix();
    glRotatef(sun_rotation, 0.0f, 0.0f, 1.0f);
    glTranslatef(8.0f, 0.0f, 0.0f);
    // nastav poziciu svetla do stredu slnka
    GLfloat sun_position[] = { 0.0f, 0.0f, 0.0f, 1.0f };
    glLightfv(GL_LIGHT0, GL_POSITION, sun_position);
    glScalef(0.3f, 0.3f, 0.3f);
    glColor3f(1.0f, 1.0f, 1.0f);
    if (selection) glLoadName(5);
    renderSphereVBO();
    glPopMatrix();
    
    // using per-pixel environment mapping
    glUseProgram(shaders_envmap);
    if (g_isTexture)
        SetShaderUniform1i(shaders_envmap, "texturing_enabled", 1);
    else
        SetShaderUniform1i(shaders_envmap, "texturing_enabled", 0);

    // vykresli 4x torus s environment mapou
    glPushMatrix();
    glLoadIdentity();
    SetShaderUniformMatrix4f(shaders_envmap, "view_matrix", view_matrix);
    SetShaderUniform1i(shaders_envmap, "color_texture", 0);
    SetShaderUniform1i(shaders_envmap, "cubemap_texture", 1);
    SetShaderUniform4f(shaders_envmap, "camera_world", camera_pos);
    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, tex_lava);
    glActiveTexture(GL_TEXTURE1);
    glBindTexture(GL_TEXTURE_CUBE_MAP, tex_skybox);
    glTranslatef(4, 0, 0);
    if (selection) glLoadName(1);
    renderTorusVBO();
    glTranslatef(-8, 0, 0);
    if (selection) glLoadName(2);
    renderTorusVBO();
    glTranslatef(4, 4, 0);
    if (selection) glLoadName(3);
    renderTorusVBO();
    glTranslatef(0, -8, 0);
    if (selection) glLoadName(4);
    renderTorusVBO();
    glBindTexture(GL_TEXTURE_CUBE_MAP, 0);
    glActiveTexture(GL_TEXTURE0);
    glPopMatrix();
    glUseProgram(0);

    // vykresli casticovy system
    if (!selection)
        renderParticles();

    // vykresli 3D text
    glEnable(GL_LIGHTING);
    glColor3f(0.0f, 1.0f, 1.0f);
    if (selection) glLoadName(6);
    render3DTextWGL(7, 0, 3, 1, "Particle system & picking");
    glDisable(GL_LIGHTING);
}

// Procedura obsluhujuca vykreslovanie
void render(void)   
{
    // Vymaz (vypln) obrazovku a z-buffer definovanymi hodnotami
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);	

    renderScene(false);

    // vykresli 2D text
    glColor3f(1.0f, 1.0f, 0.0f);
    render2DTextFT(10, g_window_height - 20, "Project OpenGL - Picking, Particle System");
    render2DTextFT(10, g_window_height - 45, "F1 - zapni/vypni animaciu");
    render2DTextFT(10, g_window_height - 70, "F2 - zapni/vypni textury");
    char string[1024];
    sprintf(string, "Picked Object: %i", g_id_selected_object);
    render2DTextFT(10, g_window_height - 95, string);

    // Zamen buffre a zobraz vykreslene objekty
    glutSwapBuffers();
}

// Procedura obsluhujuca zmenu rozmerov okna
void reshape(int w, int h)
{
    g_window_width = w;
    g_window_height = h;

    // nastvenie casti okna pre vykreslenie, tentoraz kreslime do celeho okna
	glViewport(0, 0, w, h);
    // Budeme menit projekcnu maticu (transformaciu)
	glMatrixMode(GL_PROJECTION);                        
	glLoadIdentity();                                  

	if (h == 0) 
        h = 1;
    // Chceme perspektivnu projekciu
    float ratio = (float)w/(float)h;
	gluPerspective(60, ratio, 0.1, 1000.0);
    //glOrtho(-10.0f*ratio, 10.0f*ratio, -10.0f, 10.0f, 1.0, 5000.0); 
}

// Procedura obsluhujuca stlacenie klavesy
void keyboard(unsigned char key, int x, int y)
{
	switch (key) 
    {
        // Ak Esc bolo stlacene
		case 27:
            DeleteAllVBO();
            // Ukoncenie programu
			exit(0);
		    break;
		default:        
		    break;
	}
}

// Procedura obsluhujuca stlacenie specialnej klavesy
void special_keys(int a_keys, int x, int y)
{
	switch (a_keys) 
    {
        // Ak F1 bolo stlacene
		case GLUT_KEY_F1:
            g_isAnim = !g_isAnim;
            glutPostRedisplay();
		    break;
        // Ak F2 bolo stlacene
        case GLUT_KEY_F2:
            g_isTexture = !g_isTexture;
            glutPostRedisplay();
		    break;
        // Ak F3 bolo stlacene
        case GLUT_KEY_F3:
            g_isMultisample = !g_isMultisample;
            if (g_isMultisample)
                glEnable(GL_MULTISAMPLE);
            else
                glDisable(GL_MULTISAMPLE);
            glutPostRedisplay();
		    break;
		default:
		    break;
	}
}

void processSelectionHits(GLint hits, GLuint buffer[])
{
    unsigned int i, j;
    GLuint names, *ptr;

    GLint min_val = 4294967295;
    g_id_selected_object = -1;

    printf("Number of hits = %d\n", hits);
    ptr = (GLuint *) buffer;
    for (i = 0; i < hits; i++) 
    {
        names = *ptr;
        printf ("Values for hit no. %d\n", i); 
        printf ("  number of names in stack = %d\n", names); ptr++;
        printf ("  z1 = %u\n", *ptr); GLuint z_min = GLuint(*ptr);ptr++;
        printf ("  z2 = %u\n", *ptr); ptr++;
        printf ("  the name stack: ");
        GLuint id_object = -1;
        for (j = 0; j < names; j++) 
        {
            printf ("%d, ", *ptr);
            id_object = GLuint(*ptr);
            ptr++;
        }
        printf ("\n");

        if (z_min < min_val)
        {
            min_val = z_min;
            g_id_selected_object = id_object;
        }
    }
    printf ("\n");
}

// Procedura obsluhujuca stlacenie tlacitka mysi
void mouse_klik(int button, int state, int x, int y )
{
	switch (button) 
    {
		case GLUT_LEFT_BUTTON:
            if (state == GLUT_DOWN)
            {
                left_down = true;
			    left_mouse_down_x = x;
			    left_mouse_down_y = y;
			    old_fi = fi;
			    old_xi = xi;
                break;
            }
            else if (state == GLUT_UP)
            {
                left_down = false;

                // zisti najblizsi objekt pod kurzorom
                GLint viewport[4];
                glGetIntegerv (GL_VIEWPORT, viewport);

                // prepare projection matrix with picking and view matrix
                float proj_matrix[16];
                glGetFloatv(GL_PROJECTION_MATRIX, proj_matrix);
                glMatrixMode (GL_PROJECTION);
                glPushMatrix();
                glLoadIdentity ();
                gluPickMatrix(x, viewport[3] - y, 3, 3, viewport);
                glMultMatrixf(proj_matrix);

                // reder scene in selection mode using prepared transformations
                const int BUFSIZE = 2048;
                GLuint selectBuf[BUFSIZE];
                glSelectBuffer(BUFSIZE, selectBuf);
                glRenderMode(GL_SELECT);
                renderScene(true);
                glFlush();

                glMatrixMode (GL_PROJECTION);
                glPopMatrix();
                glMatrixMode (GL_MODELVIEW);

                // return to normal mode and write all objects under cursor
                GLint hits = glRenderMode(GL_RENDER);
                processSelectionHits (hits, selectBuf);
            }
		    break;
        case GLUT_RIGHT_BUTTON:
            if (state == GLUT_DOWN)
            {
                right_down = true;
			    right_mouse_down_y = y;
			    old_vzd = vzd;
                break;
            }
            else if (state == GLUT_UP)
            {
                right_down = false;
            }
		    break;
		default:
		    break;
	}

    glutPostRedisplay();
}

// Procedura obsluhujuca pohyb mysi
void mouse_move(int x, int y)
{
    if (left_down == true)
    {
        fi = old_fi + (left_mouse_down_x - x)/200.0f;
        xi = old_xi + (y - left_mouse_down_y)/200.0f;
        glutPostRedisplay();
    }

    if (right_down == true)
    {
        vzd = old_vzd + (right_mouse_down_y - y)/10.0f;
        glutPostRedisplay();
    } 

}

void timer(int id)
{
    if (g_isAnim)
    {
        sun_rotation += 0.5f;
        updateParticles();
    }

    glutPostRedisplay();
    glutTimerFunc(50, timer, 0);
}

// Zakladna funkcia spustena na zaciatku
int main(int argc, char** argv)
{
    // GLUT Inicializacia
	glutInit(&argc, argv);                              
    // Display Mod (Rgb a Double Buffering)
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH | GLUT_STENCIL | GLUT_MULTISAMPLE);
    //glutInitDisplayString("stencil>=8 rgb>=8 double depth>=24");
    // Nastavime pociatocnu poziciu a velkost okna
    glutInitWindowPosition(50, 50);
    glutInitWindowSize(800, 600);                   
    // Vyrobime okno s danym menom
    glutCreateWindow("--== OpenGL projekt ==--" );  

    // Zinicializujeme GLEW
    glewInit();

    // Nastavime zakladne vlastnosti OpenGL
	init();
    initSphereVBO(32, 32);
    initTorusVBO(0.6f, 2.0f, 32, 32);
    initTextures();
    initShaders();
    init2DFontFT("fonts/optien.ttf");
    initParticles();

    // Zaregistrujeme callback funkcie
	glutDisplayFunc(render);                            
	glutReshapeFunc(reshape);                        
	glutKeyboardFunc(keyboard);                      
	glutSpecialFunc(special_keys);
    glutMouseFunc(mouse_klik);
    glutMotionFunc(mouse_move);
    glutTimerFunc(0, timer, 0);

    // Spustenie hlavneho okruhu zachytavania sprav
	glutMainLoop();

	return 0;
}
